from .helper_functions import check_validity_of_function_call
from .data_processing_punctuality import plot_lateness
from .data_processing_speed import (
    plot_street_percentage, 
    plot_district_percentage, 
    plot_percentage_per_district, 
    plot_number_of_infractions, 
    draw_map
)
from .data_processing_theaters import plot_theaters_map


def analyze_all(data_set: int, streets: bool = False) -> None:
    '''Analyzes the data from the given data set and plots the results'''

    # Checks whether the function call is valid, ie. whether the data
    # set exists and whether the data set has the required data
    valid_call = check_validity_of_function_call(
        data_set, streets, True)

    if not valid_call:
        raise ValueError("Invalid function call")

    plot_theaters_map(data_set)

    analyze_punctuality(data_set, False)
    analyze_speed(data_set, streets, False)


def analyze_speed(data_set: int, streets: bool = False,
                  check_validity: bool = True) -> None:
    '''Only analyzes the speed of the busses in the given data
    set and plots the results
    '''

    if check_validity:
        valid_call = check_validity_of_function_call(
            data_set, streets, False)

        if not valid_call:
            raise ValueError("Invalid function call")

    if streets:
        plot_street_percentage(data_set)

    plot_district_percentage(data_set)
    plot_percentage_per_district(data_set)
    plot_number_of_infractions(data_set)
    draw_map(data_set)


def analyze_punctuality(data_set: int, check_validity: bool = True) -> None:
    '''Only analyzes the punctuality of the busses in the given data'''

    if check_validity:
        valid_call = check_validity_of_function_call(
            data_set, False, True)

        if not valid_call:
            raise ValueError("Invalid function call")

    plot_lateness(data_set)


def analyze_theaters(data_set: int, check_validity: bool = True) -> None:
    '''Only analyzes the theaters in the given data set
    and plots the results
    '''

    if check_validity:
        valid_call = check_validity_of_function_call(
            data_set, False, False)

        if not valid_call:
            raise ValueError("Invalid function call")

    plot_theaters_map(data_set)
    


from .fetch_bus_routes import collect_bus_routes
from .fetch_current_positions import fetch_current_positions
from .fetch_bus_schedules import collect_bus_schedules
from .fetch_theaters import fetch_theaters
from .fetch_bus_stops import fetch_bus_stops


def collect_all_data(api_key, data_set_size: int = 60, spacing: int = 60) -> None:
    '''Collects all the data for the project'''

    collect_bus_schedules(api_key)
    collect_bus_routes(api_key)
    fetch_bus_stops(api_key)
    fetch_theaters(api_key)
    fetch_current_positions(api_key, data_set_size, spacing)



import json
import requests
import os


def format_response(response: dict) -> list:
    '''Formats the response from the UM API into a more readable format'''
        
    formatted_response = []

    for feature in response['result']['featureMemberList']:
        theater = {
            'name': feature['properties'][4]['value'],
            'latitude': feature['geometry']['coordinates'][0]['latitude'],
            'longitude': feature['geometry']['coordinates'][0]['longitude'],
            'address': feature['properties'][1]['value'],
            'district': feature['properties'][5]['value']
        }

        formatted_response.append(theater)

    return formatted_response


def fetch_theaters(api_key: str, path: str = 'theaters_data') -> None:
    '''Fetches the data about the theaters in Warsaw and saves 
    it to a JSON file
    '''
        
    # Set the api_key and the url for UM Warszawa
    api_key = api_key
    url = 'https://api.um.warszawa.pl/api/action/wfsstore_get'

    params = {
        'id': 'e26218cb-61ec-4ccb-81cc-fd19a6fee0f8',
        'apikey': api_key
    }

    response = requests.get(url, params=params).json()
        
    response = format_response(response)

    if not os.path.isdir('theaters_data'):
        os.mkdir('theaters_data')

    with open(f'{path}/theaters.json', 'w') as file:
        json.dump(response, file)



import json
import os
import requests


def collect_bus_routes(api_key: str, path: str = 'bus_data') -> None:
    '''Collects the bus routes and saves them in a json file'''

    url = 'https://api.um.warszawa.pl/api/action/public_transport_routes/'

    params = {
        'apikey': api_key,
        'type': '1'
    }

    # The request is repeated until it is successful
    while True:
        try:
            response = requests.get(url, params=params)
            response.raise_for_status()
            break
        except BaseException:
            continue

    file_path = os.path.join(f'{path}', 'bus_routes.json')

    with open(file_path, 'w') as file:
        json.dump(response.json(), file)



import json
import os
import time
from .helper_functions import fetch_data

# Set the url for UM Warszawa
url = 'https://api.um.warszawa.pl/api/action/busestrams_get'


def get_next_folder_number(path: str = 'data_sets') -> int:
    return len(os.listdir(f'{path}'))


def fetch_current_positions(
        api_key: str,
        iterations: int,
        spacing: int = 10,
        path: str = 'data_sets') -> None:
    
    '''Fetches data for (iterations * spacing / 60) minutes and places into a
    dict then into a json file
    '''

    # Parameters for the request
    params = {
        'resource_id': 'f2e5503e-927d-4ad3-9500-4ab9e55deb59',
        'apikey': api_key,
        'type': '1'
    }

    folder_number = get_next_folder_number(path)
    folder_path = os.path.join(f'{path}', 'data_set_' + str(folder_number))

    # Always creates a new folder for the data set
    os.makedirs(folder_path)

    i = 0

    while i < iterations:
        data = fetch_data(params, url)

        if data is None:
            continue

        # Saves the data in a dict for easier access
        response_dict = dict()

        for item in data['result']:
            response_dict[item['VehicleNumber']] = item

        file_path = os.path.join(folder_path, f'{i}.json')

        with open(file_path, 'w') as file:
            json.dump(response_dict, file)

        i += 1
        time.sleep(spacing)



import time
import warnings
import requests
import time
import json
import pkg_resources
import geopandas as gpd
import os

warnings.filterwarnings('ignore')


def fetch_data(params, url, attempts=20) -> dict:
    '''Forces the Warsaw UM API to return the data in a number of
    attempts and retries if the request fails.
    '''

    attempted = 0

    while True:
        if attempted >= attempts:
            return None

        try:
            response = requests.get(url, params=params, timeout=10)
            response.raise_for_status()
            response = response.json()

            if response['result'][0] == 'B':
                attempted += 1
                continue
            else:
                break
        except (requests.exceptions.RequestException, ValueError) as err:
            attempted += 1
            time.sleep(1)
        except BaseException:
            attempted += 1
            continue

    return response


def load_geojson(file_path: str) -> dict:
    '''Loads a geojson file from the package resources 
    and returns a dictionary
    '''

    with pkg_resources.resource_stream(__name__, file_path) as file:
        gdf = json.load(file)

    return gdf


def read_geojson(file_path: str) -> gpd.GeoDataFrame:
    '''Reads a geojson file from the package resources 
    and returns a GeoDataFrame
    '''

    with pkg_resources.resource_stream(__name__, file_path) as file:
        gdf = gpd.read_file(file)

    return gdf


def check_validity_of_function_call(
        data_set: int,
        streets: bool = False,
        lateness: bool = False) -> bool:
    '''Verifies if the function call is valid. If the data set does not exist,
    or if the data set does not have street data or lateness data, the 
    function returns False. Otherwise, it returns True.
    '''

    data_set_dir = f"data_sets/data_set_{data_set}"

    if not os.path.isdir(data_set_dir):
        print(f"Data set {data_set} does not exist.")
        return False

    if not os.path.isfile(f'filtered_data/filtered_data_{data_set}.json'):
        print(f"Data set {data_set} does not have street data.")
        return False

    if streets:
        with open(f'filtered_data/filtered_data_{data_set}.json') as file:
            data = json.load(file)
            street_1 = data[0]['Street_1']
            street_2 = data[0]['Street_2']

        if street_1 == '' or street_2 == '':
            print(f"Data set {data_set} does not have street data.")
            return False

    if lateness and not os.path.isfile(
            f'lateness_data/lateness_{data_set}.json'):
        return False

    return True


def create_directory(data_set: int) -> None:
    '''Creates a directory for the plots of a specific data 
    set if it does not exist already.
    '''

    directory = f'plots/plots_{data_set}'

    if not os.path.exists(directory):
        os.makedirs(directory)



from bus_project_NJ.src.fetch_theaters import fetch_theaters
from bus_project_NJ.src.fetch_current_positions import fetch_current_positions
from bus_project_NJ.src.fetch_bus_stops import fetch_bus_stops
from bus_project_NJ.src.fetch_bus_routes import collect_bus_routes
from bus_project_NJ.src.fetch_bus_schedules import collect_bus_schedules
import pytest
from unittest.mock import MagicMock, patch
import sys
import json
import os
import shutil
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))


# Test 1: tests fetching data regarding theaters from API
@pytest.fixture
def mock_theater_response():
    response = {
        "result": {
            "featureMemberPropertyKey": [
                "OBJECTID",
                "ULICA",
                "NUMER",
                "KOD",
                "OPIS",
                "DZIELNICA",
                "JEDN_ADM",
                "TEL_FAX",
                "WWW",
                "MAIL",
                "AKTU_DAN"
            ],
            "featureMemberList": [
                {
                    "geometry": {
                        "type": "ShapePoint",
                        "coordinates": [
                            {
                                "latitude": "52.217160",
                                "longitude": "20.989056"
                            }
                        ]
                    },
                    "properties": [
                        {
                            "value": "13",
                            "key": "OBJECTID"
                        },
                        {
                            "value": "M. Reja",
                            "key": "ULICA"
                        },
                        {
                            "value": "9",
                            "key": "NUMER"
                        },
                        {
                            "value": "02-104",
                            "key": "KOD"
                        },
                        {
                            "value": "Teatr Ochoty",
                            "key": "OPIS"
                        },
                        {
                            "value": "Ochota",
                            "key": "DZIELNICA"
                        },
                        {
                            "value": "Warszawa",
                            "key": "JEDN_ADM"
                        },
                        {
                            "value": "535 395 513",
                            "key": "TEL_FAX"
                        },
                        {
                            "value": "http:\\/\\/www.teatrochoty.pl\\/",
                            "key": "WWW"
                        },
                        {
                            "value": "mailto:sekretariat@teatrochoty.pl",
                            "key": "MAIL"
                        },
                        {
                            "value": "czerwiec 2014",
                            "key": "AKTU_DAN"
                        }
                    ]
                }
            ]
        }
    }

    return response


@pytest.fixture
def expected_theater_response():
    return [
        {
            "name": "Teatr Ochoty",
            "latitude": "52.217160",
            "longitude": "20.989056",
            "address": "M. Reja",
            "district": "Ochota"
        }
    ]


def test_fetching_theaters(mock_theater_response, expected_theater_response):
    with patch('bus_project_NJ.src.fetch_theaters.requests.get') as mock_get:
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.return_value = mock_theater_response

        folder_path = 'test'

        fetch_theaters("dummy_key", folder_path)

        with open('test/theaters.json', 'r') as file:
            data = json.load(file)
            assert data == expected_theater_response


# Test 2.1: tests fetching data regarding the current location of busses
# from API


@pytest.fixture
def mock_current_positions_1():
    return {
        "result": [
            {
                "Lines": "test1",
                "Lon": 21.041642,
                "VehicleNumber": "1000",
                "Time": "2024-02-19 19:29:29",
                "Lat": 52.24989,
                "Brigade": "1"
            },
            {
                "Lines": "test2",
                "Lon": 21.035853,
                "VehicleNumber": "1001",
                "Time": "2024-02-19 19:29:36",
                "Lat": 52.2227791,
                "Brigade": "3"
            },
            {
                "Lines": "test3",
                "Lon": 21.034853,
                "VehicleNumber": "1002",
                "Time": "2024-02-19 18:05:48",
                "Lat": 52.253721,
                "Brigade": "4"
            }
        ]
    }


@pytest.fixture
def expected_current_positions_1():
    return {
        "1000": {
            "Lines": "test1",
            "Lon": 21.041642,
            "VehicleNumber": "1000",
            "Time": "2024-02-19 19:29:29",
            "Lat": 52.24989,
            "Brigade": "1"
        },
        "1001": {
            "Lines": "test2",
            "Lon": 21.035853,
            "VehicleNumber": "1001",
            "Time": "2024-02-19 19:29:36",
            "Lat": 52.2227791,
            "Brigade": "3"
        },
        "1002": {
            "Lines": "test3",
            "Lon": 21.034853,
            "VehicleNumber": "1002",
            "Time": "2024-02-19 18:05:48",
            "Lat": 52.253721,
            "Brigade": "4"
        }
    }


def test_fetching_current_positions(
        mock_current_positions_1,
        expected_current_positions_1):

    with patch('bus_project_NJ.src.helper_functions.requests.get') as mock_get:
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.return_value = mock_current_positions_1

        file_number = len(os.listdir('test'))

        fetch_current_positions("dummy_key", 1, 10, 'test')

        with open(f'test/data_set_{file_number}/0.json', 'r') as file:
            data = json.load(file)
            assert data == expected_current_positions_1

# Test 2.2: tests fetching data regarding the current location of busses
# from API


@pytest.fixture
def mock_current_positions_2():
    return {
        "result": [
            {
                "Lines": "test1",
                "Lon": 21.033248,
                "VehicleNumber": "1000",
                "Time": "2024-02-19 19:30:36",
                "Lat": 52.260526,
                "Brigade": "1"
            },
            {
                "Lines": "test2",
                "Lon": 21.033248,
                "VehicleNumber": "1001",
                "Time": "2024-02-19 19:36:36",
                "Lat": 52.267934,
                "Brigade": "3"
            },
            {
                "Lines": "test3",
                "Lon": 21.033248,
                "VehicleNumber": "1002",
                "Time": "2024-02-19 18:07:58",
                "Lat": 52.274809,
                "Brigade": "4"
            }
        ]
    }


@pytest.fixture
def expected_current_positions_2():
    return {
        "1000": {
            "Lines": "test1",
            "Lon": 21.033248,
            "VehicleNumber": "1000",
            "Time": "2024-02-19 19:30:36",
            "Lat": 52.260526,
            "Brigade": "1"
        },
        "1001": {
            "Lines": "test2",
            "Lon": 21.033248,
            "VehicleNumber": "1001",
            "Time": "2024-02-19 19:36:36",
            "Lat": 52.267934,
            "Brigade": "3"
        },
        "1002": {
            "Lines": "test3",
            "Lon": 21.033248,
            "VehicleNumber": "1002",
            "Time": "2024-02-19 18:07:58",
            "Lat": 52.274809,
            "Brigade": "4"
        }
    }


def test_fetching_current_positions_2(
        mock_current_positions_2,
        expected_current_positions_2):

    with patch('bus_project_NJ.src.helper_functions.requests.get') as mock_get:
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.return_value = mock_current_positions_2

        file_number = len(os.listdir('test'))

        fetch_current_positions("dummy_key", 1, 10, 'test')

        with open(f'test/data_set_{file_number}/0.json', 'r') as file:
            data = json.load(file)
            assert data == expected_current_positions_2

            existing_file_path = f'test/data_set_{file_number}/0.json'

            new_file_path = f'test/data_set_{file_number - 1}/1.json'

            with open(existing_file_path, 'r') as existing_file:
                data = json.load(existing_file)

            with open(new_file_path, 'w') as new_file:
                json.dump(data, new_file)

            os.remove(existing_file_path)

            folder_path = f'test/data_set_{file_number}'
            shutil.rmtree(folder_path)

# Test 3: tests fetching data regarding bus stops from API


@pytest.fixture
def mock_bus_stops():
    return {
        "result": [
            {
                "values": [
                    {"value": "1001", "key": "zespol"},
                    {"value": "01", "key": "slupek"},
                    {"value": "Kijowska", "key": "nazwa_zespolu"},
                    {"value": "2201", "key": "id_ulicy"},
                    {"value": "52.248455", "key": "szer_geo"},
                    {"value": "21.044827", "key": "dlug_geo"},
                    {"value": "al.Zieleniecka", "key": "kierunek"},
                    {"value": "2023-10-14 00:00:00.0", "key": "obowiazuje_od"}
                ]
            },
            {
                "values": [
                    {"value": "1002", "key": "zespol"},
                    {"value": "01", "key": "slupek"},
                    {"value": "Z\u0105bkowska", "key": "nazwa_zespolu"},
                    {"value": "2201", "key": "id_ulicy"},
                    {"value": "52.251325", "key": "szer_geo"},
                    {"value": "21.038457", "key": "dlug_geo"},
                    {"value": "Kijowska", "key": "kierunek"},
                    {"value": "2023-10-14 00:00:00.0", "key": "obowiazuje_od"}
                ]
            },
            {
                "values": [
                    {"value": "1003", "key": "zespol"},
                    {"value": "03", "key": "slupek"},
                    {"value": "Dw.Wile\u0144ski", "key": "nazwa_zespolu"},
                    {"value": "0123", "key": "id_ulicy"},
                    {"value": "52.254118", "key": "szer_geo"},
                    {"value": "21.033248", "key": "dlug_geo"},
                    {"value": "Park Praski", "key": "kierunek"},
                    {"value": "2023-10-14 00:00:00.0", "key": "obowiazuje_od"}
                ]
            }
        ]
    }


@pytest.fixture
def expected_bus_stops_response():
    return {
        "1001-01": {
            "zespol": "1001",
            "slupek": "01",
            "nazwa_zespolu": "Kijowska",
            "id_ulicy": "2201",
            "szer_geo": "52.248455",
            "dlug_geo": "21.044827",
            "kierunek": "al.Zieleniecka",
            "obowiazuje_od": "2023-10-14 00:00:00.0"
        },
        "1002-01": {
            "zespol": "1002",
            "slupek": "01",
            "nazwa_zespolu": "Ząbkowska",
            "id_ulicy": "2201",
            "szer_geo": "52.251325",
            "dlug_geo": "21.038457",
            "kierunek": "Kijowska",
            "obowiazuje_od": "2023-10-14 00:00:00.0"
        },
        "1003-03": {
            "zespol": "1003",
            "slupek": "03",
            "nazwa_zespolu": "Dw.Wileński",
            "id_ulicy": "0123",
            "szer_geo": "52.254118",
            "dlug_geo": "21.033248",
            "kierunek": "Park Praski",
            "obowiazuje_od": "2023-10-14 00:00:00.0"
        }
    }


def test_fetching_bus_stops(mock_bus_stops, expected_bus_stops_response):
    with patch('bus_project_NJ.src.helper_functions.requests.get') as mock_get:
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.return_value = mock_bus_stops

        folder_path = 'test'

        fetch_bus_stops("dummy_key", folder_path)

        with open('test/bus_stops.json', 'r') as file:
            data = json.load(file)
            assert data == expected_bus_stops_response

# Test 4: tests fetching data regarding bus routes from API


@pytest.fixture
def mock_bus_routes():
    return {
        "result": {
            "test1": {
                "test_route_1": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            },
            "test2": {
                "test_route_2": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            },
            "test3": {
                "test_route_3": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            }
        }
    }


@pytest.fixture
def expected_bus_routes_response():
    return {
        "result": {
            "test1": {
                "test_route_1": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            },
            "test2": {
                "test_route_2": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            },
            "test3": {
                "test_route_3": {
                    "11": {
                        "odleglosc": 7033,
                        "ulica_id": "2201",
                        "nr_zespolu": "1001",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "10": {
                        "odleglosc": 6393,
                        "ulica_id": "2201",
                        "nr_zespolu": "1002",
                        "typ": "1",
                        "nr_przystanku": "01"
                    },
                    "12": {
                        "odleglosc": 7613,
                        "ulica_id": "0123",
                        "nr_zespolu": "1003",
                        "typ": "1",
                        "nr_przystanku": "03"
                    }
                }
            }
        }
    }


def test_fetching_bus_routes(mock_bus_routes, expected_bus_routes_response):
    with patch('bus_project_NJ.src.helper_functions.requests.get') as mock_get:
        mock_get.return_value = MagicMock(status_code=200)
        mock_get.return_value.json.return_value = mock_bus_routes

        folder_path = 'test'

        collect_bus_routes("dummy_key", folder_path)

        with open('test/bus_routes.json', 'r') as file:
            data = json.load(file)
            assert data == expected_bus_routes_response


import json
import os
import warnings
import geopy.distance
import requests
from .helper_functions import load_geojson
from typing import Tuple
from datetime import datetime
from shapely.geometry import Point, shape

warnings.filterwarnings('ignore')

RADIUS = 700 # The radius in meters for the GUGiK API request


def dump_filtered_data(data_set: int, 
                       filtered_data: list, 
                       path: str = 'filtered_data') -> None:
    with open(os.path.join(f'{path}',
                           f"filtered_data_{data_set}.json"), 'w') as outfile:
        json.dump(filtered_data, outfile)


def calculate_velocity(pos_1: tuple, pos_2: tuple, time_1, time_2) -> float:
    '''Calculates the velocity of a vehicle between two points in km/h'''

    try:
        distance = geopy.distance.distance(pos_1, pos_2).km
    except BaseException:
        return None

    time_difference = (time_2 - time_1).seconds

    if (time_difference == 0):
        velocity = 0
    else:
        velocity = distance / time_difference * 3600

    return velocity


def determine_district(point_coords: tuple) -> str:
    '''Determines the district of a given point using the geojson file
    warszawa-dzielnice.geojson; returns None if the point is not
    in any district
    '''

    districts = load_geojson('warszawa-dzielnice.geojson')

    for feature in districts['features'][1:]:
        district_shape = shape(feature['geometry'])
        if district_shape.contains(Point(point_coords)):
            return feature['properties']['name']
    return None


def determine_street(pos: tuple) -> str:
    '''Determines the street of a given point using the GUGiK API;
    returns None if the request fails
    '''

    position = "POINT(" + str(pos[0]) + " " + str(pos[1]) + ")"

    try:
        street = requests.post(
            f"http://services.gugik.gov.pl/uug/?request=GetAddressReverse&location={position}&srid=4326&radius={RADIUS}").json()['results']['1']['street']
    except BaseException:
        return None

    return street


def valid_position(pos: tuple) -> bool:
    '''Checks if the position is valid'''

    if (pos[0] < -180 or pos[0] > 180 or pos[1] < -90 or pos[1] > 90):
        return False

    return True


def filter_files(
        file_1_dict: dict,
        file_2_dict: dict,
        filtered_data: list,
        streets: bool) -> Tuple[int, int]:
    '''Filters the data from two files, removing any data that is
    not useful for the analysis or is invalid
    '''

    all = 0
    rejected = 0

    for vehicle_number in file_1_dict:
        all += 1

        if vehicle_number not in file_2_dict:
            rejected += 1
            continue

        pos_1 = (file_1_dict[vehicle_number]['Lon'],
                 file_1_dict[vehicle_number]['Lat'])
        pos_2 = (file_2_dict[vehicle_number]['Lon'],
                 file_2_dict[vehicle_number]['Lat'])

        # If the position is invalid, because the latitude or longitude is out
        # of range, skip the entry
        if not valid_position(pos_1) or not valid_position(pos_2):
            rejected += 1
            continue

        try:
            # Convert the time to a datetime object. If the time is invalid,
            # skip the entry
            time_1 = datetime.strptime(
                file_1_dict[vehicle_number]['Time'], '%Y-%m-%d %H:%M:%S')
            time_2 = datetime.strptime(
                file_2_dict[vehicle_number]['Time'], '%Y-%m-%d %H:%M:%S')
        except BaseException:
            rejected += 1
            continue

        velocity = calculate_velocity(pos_1, pos_2, time_1, time_2)

        # Get the district of the start and end positions
        district_start = determine_district(
            (file_1_dict[vehicle_number]['Lon'],
             file_1_dict[vehicle_number]['Lat']))

        district_end = determine_district(
            (file_2_dict[vehicle_number]['Lon'],
             file_2_dict[vehicle_number]['Lat']))

        if district_start is None or velocity is None or district_end is None:
            rejected += 1
            continue

        if streets:
            # Get the street of the start and end positions
            street_1 = determine_street(pos_1)
            street_2 = determine_street(pos_2)

            # If the request to the GUGiK API fails, skip the entry
            if street_1 is None or street_2 is None:
                rejected += 1
                continue
        else:
            street_1 = ""
            street_2 = ""

        data = {
            "Start_pos": pos_1,
            "End_pos": pos_2,
            "District_Start": district_start,
            "District_End": district_end,
            "Velocity": velocity,
            "Line": file_1_dict[vehicle_number]['Lines'],
            "Street_1": street_1,
            "Street_2": street_2,
            "Vehicle_number": vehicle_number,
            "Brigade": file_1_dict[vehicle_number]['Brigade'],
            "Time_1": file_1_dict[vehicle_number]['Time'],
            "Time_2": file_2_dict[vehicle_number]['Time']
        }

        filtered_data.append(data)

    return (all, rejected)


def filter_data_set(
        data_set: int,
        streets: bool = False,
        radius: int = 700,
        path: str = 'filtered_data',
        src_path: str = 'data_sets') -> None:
    '''Filters the data from the selected data set, by filtering each two
    consecutive files and removing invalid data and removes the invalid
    data from the data set
    '''
    
    global RADIUS
    RADIUS = radius

    all = 0
    rejected = 0

    filtered_data = []

    subfolder_path = os.path.join(f'{src_path}', f"data_set_{data_set}")

    number_of_files = len(os.listdir(subfolder_path))

    file_path_1 = os.path.join(subfolder_path, '0.json')
    open_file_1 = open(file_path_1)
    file_1_dict = json.load(open_file_1)

    for i in range(number_of_files - 1):
        print("\tWORKING ON FILE:", i + 1)

        file_path_2 = os.path.join(subfolder_path, f'{i + 1}.json')
        open_file_2 = open(file_path_2)
        file_2_dict = json.load(open_file_2)

        stats = filter_files(file_1_dict, file_2_dict, filtered_data, streets)

        all += stats[0]
        rejected += stats[1]

        open_file_1.close()
        file_path_1 = file_path_2
        file_1_dict = file_2_dict
        open_file_1 = open_file_2

    open_file_2.close()

    dump_filtered_data(data_set, filtered_data, path)

    print(
        f"Rejected {rejected} out of {all} entries which \
        is {round(rejected / all * 100, 2)}%")



from bus_project_NJ.src.filter_data import filter_data_set
import pytest
from unittest.mock import MagicMock, patch
import sys
import json
import os
import shutil
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

# Test 1: test the filtering of data (their velocity and other data)
def test_filter_data_set():
    expected_filtered_data = [
        {
            "Start_pos": [21.041642, 52.24989],
            "End_pos": [21.033248, 52.260526],
            "District_Start": "Praga Północ",
            "District_End": "Praga Północ",
            "Velocity": 77.60344478474863,
            "Line": "test1",
            "Street_1": "",
            "Street_2": "",
            "Vehicle_number": "1000",
            "Brigade": "1",
            "Time_1": "2024-02-19 19:29:29",
            "Time_2": "2024-02-19 19:30:36"
        },
        {
            "Start_pos": [21.035853, 52.2227791],
            "End_pos": [21.033248, 52.267934],
            "District_Start": "Śródmieście",
            "District_End": "Praga Północ",
            "Velocity": 40.30752757142837,
            "Line": "test2",
            "Street_1": "",
            "Street_2": "",
            "Vehicle_number": "1001",
            "Brigade": "3",
            "Time_1": "2024-02-19 19:29:36",
            "Time_2": "2024-02-19 19:36:36"
        },
        {
            "Start_pos": [21.034853, 52.253721],
            "End_pos": [21.033248, 52.274809],
            "District_Start": "Praga Północ",
            "District_End": "Targówek",
            "Velocity": 60.90151172737643,
            "Line": "test3",
            "Street_1": "",
            "Street_2": "",
            "Vehicle_number": "1002",
            "Brigade": "4",
            "Time_1": "2024-02-19 18:05:48",
            "Time_2": "2024-02-19 18:07:58"
        }
    ]
    
    filter_data_set(6, False, radius=700, path='test')
    
    with open('test/filtered_data_6.json', 'r') as f:
        data = json.load(f)
        
    assert data == expected_filtered_data
    


import matplotlib.pyplot as plt
import json
import os
import geopandas as gpd
import geopy.distance
from .helper_functions import load_geojson
from typing import Tuple
from shapely.geometry import Point, shape
from datetime import datetime

# 50 meters - the maximum distance from the bus stop to the
# bus route to be considered in the analysis
DISTANCE = 0.05


def calculate_distance(pos_1: tuple, pos_2: tuple) -> float:
    return geopy.distance.distance(pos_1, pos_2).km


def determine_district(point_coords: tuple) -> str:
    '''Determines the district of a given point using the
    warszawa-dzielnice.geojson geojson file
    '''

    districts = load_geojson('warszawa-dzielnice.geojson')

    for feature in districts['features'][1:]:
        district_shape = shape(feature['geometry'])
        if district_shape.contains(Point(point_coords)):
            return feature['properties']['name']
    return None


def get_appropriate_time(schedule: list, time: str) -> str:
    '''Returns the closest scheduled time to the given time based on the
    bus schedule and the time at which the bus sent out its location
    '''

    if int(time[11:13]) >= 24:  # Accounts for faulty data
        time = '00' + time[13:]
    else:
        time = time[11:19]

    if int(schedule[0][:2]) >= 24:
        schedule[0] = '00' + schedule[0][2:]

    # Initializes the minimum difference to the first scheduled time
    min_diff = abs(
        datetime.strptime(
            time,
            '%H:%M:%S') -
        datetime.strptime(
            schedule[0],
            '%H:%M:%S'))

    min_time = schedule[0]

    for scheduled_time in schedule:
        if int(scheduled_time[:2]) >= 24:
            scheduled_time = '00' + scheduled_time[2:]

        diff = abs(
            datetime.strptime(
                time,
                '%H:%M:%S') -
            datetime.strptime(
                scheduled_time,
                '%H:%M:%S'))

        if diff < min_diff:
            min_diff = diff
            min_time = scheduled_time

    return min_time


def get_lateness(time_1: str, time_2: str, schedule: list) -> int:
    '''Returns the lateness of the bus in seconds'''

    appropriate_time = get_appropriate_time(schedule, time_1)

    time_1 = datetime.strptime(time_1[11:19], '%H:%M:%S')
    time_2 = datetime.strptime(time_2[11:19], '%H:%M:%S')

    appropriate_time = datetime.strptime(appropriate_time, '%H:%M:%S')

    # Calculates the lowest number of seconds the bus was late to the stop
    return min((abs(appropriate_time - time_1)).seconds,
               (abs(appropriate_time - time_2)).seconds)


def open_files(data_set: int) -> Tuple[list, dict, dict]:
    '''Opens the necessary files for the analysis'''

    try:
        with open('bus_data/bus_routes.json', 'r') as file:
            bus_routes = json.load(file)
    except FileNotFoundError:
        print("ERROR: Bus routes file not found. "
              "Please run fetch_bus_routes.py first.")
        return None, None, None

    try:
        with open(f'filtered_data/filtered_data_{data_set}.json', 'r') as f:
            filtered_data = json.load(f)
    except FileNotFoundError:
        print(f"ERROR: Filtered data set {data_set} not found. "
              "Please run filter_data.py first.")
        return None, None, None

    try:
        with open('bus_data/bus_stops.json', 'r') as file:
            bus_stops_coordinates = json.load(file)
    except FileNotFoundError:
        print("ERROR: Bus stops file not found. \
                Please run fetch_bus_stops.py first.")
        return None, None, None

    return filtered_data, bus_routes, bus_stops_coordinates


def filter_route(
        my_bus_routes: dict,
        my_schedule: dict,
        brigade: str,
        info: dict,
        lateness: list,
        lateness_per_district: dict,
        bus_stops_coordinates: dict,
        route: list) -> None:
    '''Filters the route of the bus and adds the lateness of the bus
    to each stop and district to the respective lists and dictionaries
    '''

    for stop in my_bus_routes[route]:
        stop_id = my_bus_routes[route][stop]['nr_zespolu'] + \
            '-' + my_bus_routes[route][stop]['nr_przystanku']

        if stop_id not in bus_stops_coordinates:
            continue

        stop_coordinates = (
            bus_stops_coordinates[stop_id]['dlug_geo'],
            bus_stops_coordinates[stop_id]['szer_geo'])

        # If the bus was not close enough to the stop, it is not considered in
        # the analysis
        if (
            calculate_distance(
                info['Start_pos'],
                stop_coordinates) > DISTANCE or calculate_distance(
                info['End_pos'],
                stop_coordinates) > DISTANCE):
            continue

        # If the brigade is not in the schedule or the stop is not in the
        # schedule, it is not considered in the analysis as it is faulty data
        if brigade not in my_schedule or stop_id not in my_schedule[brigade]:
            continue

        time_late = get_lateness(
            info['Time_1'],
            info['Time_2'],
            my_schedule[brigade][stop_id])

        # If the bus was late for more than 40 minutes, it is not considered
        # in the analysis as this implies a mistake in the calculations
        if time_late >= 2400:
            continue

        lateness.append(time_late)

        district = determine_district(stop_coordinates)

        if district not in lateness_per_district:
            lateness_per_district[district] = []

        lateness_per_district[district].append(time_late)


def calculate_bus_lateness(
        data_set: int,
        bus_number: str,
        brigade: str,
        line: str) -> Tuple[list, dict]:
    '''Returns the total number of seconds that the bus line was late to each
    stop in a list and the total number of seconds that the bus line was late
    to each district in a dictionary
    '''

    filtered_data, bus_routes, bus_stops_coordinates = open_files(data_set)

    if (filtered_data is None
        or bus_routes is None
            or bus_stops_coordinates is None):
        return [], {}

    schedule_file_path = f'bus_data/bus_lines/bus{line}_schedule.json'

    if os.path.exists(schedule_file_path):
        with open(f'bus_data/bus_lines/bus{line}_schedule.json', 'r') as file:
            my_schedule = json.load(file)
    else:
        return [], {}

    my_bus_routes = bus_routes['result'][line]

    lateness = []
    lateness_per_district = {}

    for info in filtered_data:
        if info['Vehicle_number'] != bus_number or info['Brigade'] != brigade:
            continue

        for route in my_bus_routes:
            filter_route(
                my_bus_routes,
                my_schedule,
                brigade,
                info,
                lateness,
                lateness_per_district,
                bus_stops_coordinates,
                route)

    # Returns all the minutes the busses were late to each stop and
    # the total number of minutes the busses were late to each district
    return lateness, lateness_per_district


def get_lines(data_set: int) -> set:
    '''Returns the set of intervals (lines) that are present in the
    filtered data
    '''

    try:
        with open(f'filtered_data/filtered_data_{data_set}.json', 'r') as file:
            filtered_data = json.load(file)
    except FileNotFoundError:
        print(
            f"ERROR: Filtered data set {data_set} not found. "
            f"Please run filter_data.py first.")
        return None

    lines = set()

    for info in filtered_data:
        lines.add((info['Line'], info['Brigade'], info['Vehicle_number']))

    return lines


def calculate_total_lateness(data_set: int, path: str = 'lateness_data') -> None:
    '''Filters all the data in the already filtered data set and saves the
    statistics regarding how late the busses were to each stop in a json file
    '''

    total_lateness = []
    lateness_per_district = {}
    lines_lateness = {}

    lines = get_lines(data_set)

    if lines is None:
        return
    
    iter = 0

    for line in lines:
        lateness = calculate_bus_lateness(data_set, line[2], line[1], line[0])

        total_lateness.extend(lateness)
        lines_lateness[iter] = (line, lateness[0])
        
        iter += 1   

        for district in lateness[1]:
            if district not in lateness_per_district:
                lateness_per_district[district] = []

            lateness_per_district[district].extend(lateness[1][district])

    with open(f'{path}/lateness_{data_set}.json', 'w') as file:
        json.dump(lateness_per_district, file)
        
    with open(f'{path}/lateness_lines_{data_set}.json', 'w') as file:
        json.dump(lines_lateness, file)




from .src.helper_functions import check_validity_of_function_call

from .src.analyze import analyze_speed, analyze_punctuality, analyze_theaters

from .src.collect_data import collect_all_data

from .src.fetch_bus_routes import collect_bus_routes

from .src.fetch_bus_schedules import collect_bus_schedules

from .src.fetch_bus_stops import fetch_bus_stops

from .src.fetch_current_positions import fetch_current_positions

from .src.fetch_theaters import fetch_theaters

from .src.filter_data import filter_data_set

from .src.filter_data_punctuality import calculate_total_lateness

from .src.data_processing_speed import ( 
    plot_street_percentage, 
    plot_district_percentage, 
    plot_percentage_per_district, 
    plot_number_of_infractions, 
    draw_map
)

from .src.data_processing_punctuality import plot_lateness

from .src.data_processing_punctuality import plot_speed_versus_delays

from .src.data_processing_theaters import plot_theaters_map



import matplotlib.pyplot as plt
import json
import geopandas as gpd
import geopy.distance
from .helper_functions import create_directory, read_geojson
from typing import Tuple
from shapely.geometry import LineString, Point

# Set the font styles for the plots to ensure consistency and cohesion
fontstyle_annotation = {
    'fontsize': 10,
    'fontweight': 'bold',
    'color': '#235A6C'}

fontstyle_title = {'fontsize': 13, 'fontweight': 'bold', 'color': '#14333D'}

DISTANCE = 0.3
SPEED_LIMIT = 50


def calculate_distance(pos_1: tuple, pos_2: tuple) -> float:
    return geopy.distance.distance(pos_1, pos_2).km


def get_lines(data_set: int, theater_coords: tuple, path: str) -> Tuple[list, list]:
    '''Returns the intervals in which the busses were going over 
    50km/h around theaters and in general
    '''

    with open(f'{path}/filtered_data_{data_set}.json', 'r') as file:
        data = json.load(file)

    lines_theater = []
    lines = []

    for info in data:
        if (
            info['Velocity'] > 50 and calculate_distance(
                (info['Start_pos'][0],
                 info['Start_pos'][1]),
                theater_coords) < DISTANCE or calculate_distance(
                (info['End_pos'][0],
                 info['End_pos'][1]),
                theater_coords) < DISTANCE):
            lines_theater.append((info["Start_pos"], info["End_pos"]))
        elif info['Velocity'] > SPEED_LIMIT:
            lines.append((info["Start_pos"], info["End_pos"]))

    return lines_theater, lines


def configure_plot(map: plt.Axes, percentage: float) -> None:
    '''Sets the title, legend and annotation for the map of Warsaw with the 
    speeding busses and the percentage of speeding around theaters
    '''

    map.set_title(
        "Map of speeding busses in Warsaw and speeding around theaters",
        fontdict=fontstyle_title,
        y=1.05)

    map.legend(["Speeding busses", "Speeding busses around theaters"],
               facecolor='white', edgecolor='black')
    legend = map.get_legend()

    # Set color for "Speeding busses"
    legend.legendHandles[0].set_color('#96C9DC')

    # Set color for "Speeding around theaters"
    legend.legendHandles[1].set_color('red')

    map.text(
        0.96,
        0.85,
        f'{percentage}% of speeding\nhappened around theaters',
        horizontalalignment='right',
        verticalalignment='center',
        transform=map.transAxes,
        fontsize=12,
        fontdict=fontstyle_annotation)


def plot_theaters_map(data_set: int, 
                      theaters_path: str = 'theaters_data', 
                      destination: str = 'plots',
                      dump: bool = False,
                      dump_path: str = '',
                      filtered_path: str = 'filtered_data') -> None:
    '''Plots the map of Warsaw with the speeding busses 
    and the theaters marked on it
    '''

    create_directory(data_set)

    try: 
        with open(f'{theaters_path}/theaters.json', 'r') as file:
            theaters_data = json.load(file)
    except FileNotFoundError:
        print("The file theaters.json does not exist. "
              "Please fetch the theater data first.")
        return

    lines_theater = []
    lines = []

    # Plot the map of Warsaw using the GeoPandas library and the
    # warszawa-dzielnice.geojson file
    gdf = read_geojson('warszawa-dzielnice.geojson')

    map = gdf.plot(figsize=(18, 10), edgecolor="black",
                   color="#FFDDD2", legend=True, linewidth=1)

    for theater in theaters_data:
        longitude = theater['longitude']
        latitude = theater['latitude']

        point = Point(longitude, latitude)
        gdf_point = gpd.GeoDataFrame(geometry=[point], crs="EPSG:4326")
        gdf_point.plot(ax=map, color="blue", markersize=50)

        lines_theater.extend(get_lines(data_set, (longitude, latitude), f'{filtered_path}')[0])
        lines.extend(get_lines(data_set, (longitude, latitude), f'{filtered_path}')[1])
        
    if dump:
        try:
            with open(f'{dump_path}/theater_lines_{data_set}.json', 'w') as file:
                json.dump((lines_theater, lines), file)
        except FileNotFoundError:
            print(f"ERROR: File {dump_path}/theater_lines_{data_set}.json does not exist")
        return

    # Plot the lines of the speeding busses
    lines = gpd.GeoDataFrame(
        geometry=[LineString([line[0], line[1]]) for line in lines])
    lines = lines.set_crs(epsg=4326)

    lines.plot(ax=map, color="#96C9DC", linewidth=1.5, alpha=0.7, legend=True)

    lines_theater = gpd.GeoDataFrame(
        geometry=[LineString([line[0], line[1]]) for line in lines_theater])
    lines_theater = lines_theater.set_crs(epsg=4326)

    lines_theater.plot(ax=map, color="red", linewidth=1.5, legend=True)

    percentage = round(len(lines_theater) / len(lines) * 100, 2)

    configure_plot(map, percentage)

    plt.get_current_fig_manager().full_screen_toggle()
    plt.savefig(f'{destination}/plots_{data_set}/Theater_plot_{data_set}.png')


import pytest
from bus_project_NJ.src.data_processing_speed import prepare_data
from bus_project_NJ.src.data_processing_theaters import plot_theaters_map
from unittest.mock import MagicMock, patch
import sys
import json
import shutil
import os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

# Test 1: tests the data preparation for the speed analysis.
# All other functions are based upon this data directly


def test_speed_analysis():
    expected_data = [[[[21.041642,
                        52.24989],
                       [21.033248,
                        52.260526]],
                      [[21.034853,
                        52.253721],
                       [21.033248,
                        52.274809]]],
                     {"Ochota_total": 0,
                      "Rembert\u00f3w_total": 0,
                      "Bemowo_total": 0,
                      "Bia\u0142o\u0142\u0119ka_total": 0,
                      "Weso\u0142a_total": 0,
                      "Bielany_total": 0,
                      "Mokot\u00f3w_total": 0,
                      "Praga Po\u0142udnie_total": 0,
                      "Praga P\u00f3\u0142noc_total": 3,
                      "\u015ar\u00f3dmie\u015bcie_total": 1,
                      "Targ\u00f3wek_total": 1,
                      "Ursus_total": 0,
                      "Wola_total": 0,
                      "Ursyn\u00f3w_total": 0,
                      "Wawer_total": 0,
                      "Wilan\u00f3w_total": 0,
                      "W\u0142ochy_total": 0,
                      "\u017boliborz_total": 0,
                      "Praga P\u00f3\u0142noc": 2,
                      "Targ\u00f3wek": 1},
                     [{},
                      5]]

    prepare_data(6, 'test', dump=True, path='test')

    with open('test/data_dump_6.json', 'r') as f:
        data = json.load(f)

    assert data == expected_data

# Test 2: tests the data for the theater analysis


def test_theater_data():
    expected_data = [[],
                     [[[21.041642, 52.24989], [21.033248, 52.260526]],
                      [[21.034853, 52.253721], [21.033248, 52.274809]]]]

    plot_theaters_map(
        6,
        'test',
        'test',
        dump=True,
        dump_path='test',
        filtered_path='test')

    with open('test/theater_lines_6.json', 'r') as f:
        data = json.load(f)

    assert data == expected_data



import json
import os
from .helper_functions import fetch_data

# Set the url for UM Warszawa
api_key = ''
url = 'https://api.um.warszawa.pl/api/action/dbstore_get'

# Parameters for the request
params = {
    'apikey': api_key,
    'id': 'ab75c33d-3a26-4342-b36a-6e5fef0a3ac3'
}


def format_response(data: dict) -> tuple:
    '''Formats response into a more programmer-friendly format'''
    formatted_data = dict()

    for element in data['values']:
        formatted_data[element['key']] = element['value']

    key = str(formatted_data['zespol']) + "-" + str(formatted_data['slupek'])

    return (formatted_data, key)


def fetch_bus_stops(apikey: str, path: str = 'bus_data') -> None:
    '''Fetches bus stops from UM Warszawa and saves them in a json file'''
    global api_key
    api_key = apikey

    response = fetch_data(params, url, 10000)

    bus_stops_dict = dict()

    for data in response['result']:
        formatted_data = format_response(data)
        bus_stops_dict[formatted_data[1]] = formatted_data[0]

    file_path = os.path.join(f'{path}', 'bus_stops.json')

    with open(file_path, 'w') as file:
        json.dump(bus_stops_dict, file)



import matplotlib.pyplot as plt
import os
import json
import geopandas as gpd
from .helper_functions import (
    create_directory, 
    read_geojson, 
    check_validity_of_function_call, 
    load_geojson
)
from typing import Tuple
from shapely.geometry import LineString

SPEED_LIMIT = 50

# Set the font styles for the plots to ensure consistency and cohesion
fontstyle_annotation = {
    'fontsize': 10,
    'fontweight': 'bold',
    'color': '#14333D'}

fontstyle_title = {'fontsize': 13, 'fontweight': 'bold', 'color': '#14333D'}


def get_district_percentage(districts) -> dict:
    '''Returns the percentage of infractions in each district'''
    district_percentages = dict()

    for district in districts:
        if not district.endswith("_total"):
            district_percentages[district] = districts[district] / \
                districts[district + "_total"]

    return district_percentages


def draw_map(data_set: int, path: str = 'plots') -> None:
    '''Draws a map of Warsaw with the speeding busses and alongside 
    the percentage of infractions in each district. Red lines 
    represent the places where the busses were speeding
    '''
    check_validity_of_function_call(data_set)
    create_directory(data_set)

    data = prepare_data(data_set)
    lines = data[0]
    districts = data[1]

    district_percentages = get_district_percentage(districts)

    # Plot the map of Warsaw using the GeoPandas library and the
    # warszawa-dzielnice.geojson file
    gdf = read_geojson('warszawa-dzielnice.geojson')

    map = gdf.plot(
        figsize=(18, 10),
        edgecolor="black",
        color="#FFDDD2",
        legend=True,
        linewidth=1)

    map.set_title("Map of speeding busses in Warsaw", fontdict=fontstyle_title)
    map.set_xlabel("Longitude", fontsize=15, color="black")
    map.set_ylabel("Latitude", fontsize=15, color="black")

    # Plot the lines of the speeding busses
    lines = gpd.GeoDataFrame(
        geometry=[LineString([line[0], line[1]]) for line in lines])
    lines = lines.set_crs(epsg=4326)

    lines.plot(ax=map, color="red", linewidth=1.5, legend=True)

    # Plot the legend for the percentage of infractions in each district
    for district in district_percentages:
        map.scatter(
            [],
            [],
            label=f'{district} - {round(district_percentages[district] * 100, 2)}%',
            color='white')

    map.legend(title="Percentage of\nspeeding busses in each district",
               loc='center left',
               bbox_to_anchor=(1, 0.5),
               fontsize=10,
               title_fontsize=10,
               labelspacing=0.5,
               edgecolor='#14333D',
               facecolor='white',
               frameon=True,
               handlelength=0)

    plt.get_current_fig_manager().full_screen_toggle()

    plt.savefig(
        os.path.join(
            f'{path}/plots_{data_set}',
            f'Warsaw_map_{data_set}.png'))


def plot_number_of_infractions(data_set: int, path: str = 'plots') -> None:
    '''Plots the number of infractions that happened in each district'''
    check_validity_of_function_call(data_set)
    
    create_directory(data_set)

    districts = prepare_data(data_set)[1]
    districts = {district: count for district,
                 count in districts.items() if not district.endswith("_total")}

    fig, bar_chart = plt.subplots(figsize=(20, 12))

    sorted_districts = dict(
        sorted(
            districts.items(),
            key=lambda x: x[1],
            reverse=True))

    a = bar_chart.bar(
        list(
            sorted_districts.keys()), list(
            sorted_districts.values()), color='#00A8B8')

    plt.xticks(rotation=45, fontsize=10)
    plt.xlabel("District", fontdict=fontstyle_annotation)

    plt.yticks(fontsize=8)
    plt.ylabel("Number of infractions", fontdict=fontstyle_annotation)

    plt.title("Number of infractions in each district",
              fontdict=fontstyle_title)

    plt.savefig(
        os.path.join(
            f'{path}/plots_{data_set}',
            f'Number_of_infractions_{data_set}.png'))


def plot_percentage_per_district(data_set: int, path: str = 'plots') -> None:
    '''Plots the percentage of infractions that happened in each 
    district as a part of the total number of infractions in a pie chart
    '''
    check_validity_of_function_call(data_set)
    
    create_directory(data_set)

    districts = prepare_data(data_set)[1]

    total_infractions = sum([count for district, count in districts.items()
                            if not district.endswith("_total")])

    district_percentages = {
        district: (
            count
            / total_infractions)
        * 100 for district,
        count in districts.items() if not district.endswith("_total")}

    fig, pie_chart = plt.subplots(figsize=(10, 10))

    pie_chart.pie(
        list(
            district_percentages.values()),
        labels=list(
            district_percentages.keys()),
        autopct='%1.1f%%',
        pctdistance=0.85,
        colors=['#83C5BE', '#006D77', '#FFDDD2', '#E29578', '#24596B'])

    pie_chart.set_title(
        "Percentage of infractions committed in each district\n \
        as a part of the total number of infractions",
        fontdict=fontstyle_title)

    pie_chart.text(0, -1.3, "Total Infractions: {}".format(total_infractions),
                   fontsize=12, ha='center')

    plt.savefig(
        os.path.join(
            f'{path}/plots_{data_set}',
            f'Percentage_per_district_{data_set}.png'))


def plot_district_percentage(data_set: int, 
                             dst_path: str = 'plots') -> None:
    '''Plots the percentage of infractions that happened in each 
    district in a bar chart
    '''
    check_validity_of_function_call(data_set)
    
    create_directory(data_set)

    districts = prepare_data(data_set)[1]

    district_percentages = get_district_percentage(districts)

    fig, bar_chart = plt.subplots(figsize=(20, 8))

    plt.subplots_adjust(left=0.1, bottom=0.2, right=0.9,
                        top=0.85)  # Add padding

    plt.xticks(rotation=45, fontsize=10)
    plt.xlabel("District", fontdict=fontstyle_annotation)
    plt.yticks(fontsize=8)
    plt.ylabel(
        "Percentage of infractions in each district",
        fontdict=fontstyle_annotation)
    
    plt.title(
        "The percentage of speeding busses in each district",
        fontdict=fontstyle_title)

    a = bar_chart.bar(
        list(district_percentages.keys()),
        sorted(list(district_percentages.values()),reverse=True),
        color='#00A8B8')

    plt.savefig(
        os.path.join(
            f'{dst_path}/plots_{data_set}',
            f'District_percentage_{data_set}.png'))


def plot_street_percentage(data_set: int, 
                           number: int = 10, 
                           src_path: str = 'filtered_data', 
                           dst_path: str = 'plots') -> None:
    '''Plots the percentage of infractions that happened on the top
    {number} streets in a bar chart
    '''
    check_validity_of_function_call(data_set, True)
    
    create_directory(data_set)

    streets_tuple = prepare_data(data_set, src_path, streets_on=True)[2]

    if streets_tuple is None:
        return

    streets = streets_tuple[0]
    total = streets_tuple[1]

    street_percentages = {
        street: (
            count 
            / total) 
        * 100 for street, count in streets.items()}

    top_10_streets = dict(
        sorted(
            street_percentages.items(),
            key=lambda x: x[1],
            reverse=True)[:number])

    fig, bar_chart = plt.subplots(figsize=(20, 10))

    plt.xticks(rotation=45, fontsize=10, ha='right')
    plt.xlabel("Street name", fontdict=fontstyle_annotation)

    plt.yticks(fontsize=8)
    plt.ylabel(
        "Percentage of infractions on each street",
        fontdict=fontstyle_annotation)

    plt.title(
        f"The top {number} streets with the highest percentage "
        f"of speeding busses",
        fontdict=fontstyle_title)

    a = bar_chart.bar(
        list(top_10_streets.keys()), 
        list(top_10_streets.values()), color='#00A8B8')

    # Add padding
    plt.subplots_adjust(left=0.2, bottom=0.25, right=0.8, top=0.9)

    plt.savefig(
        os.path.join(
            f'{dst_path}/plots_{data_set}',
            f'Street_percentage_{data_set}.png'))


def increase_dict_values(dictionary: dict, value_1: str, value_2: str) -> None:
    '''Increases the value of the keys in the dictionary by 1. 
    If the key does not exist, it is created and set to 1. If 
    value_2 is different from value_1, it is also increased by 1.
    '''
    if value_1 not in dictionary:
        dictionary[value_1] = 0

    dictionary[value_1] += 1

    # Prevents the double counting of the same street or
    # district in the preparation of the data
    if value_2 != value_1:
        if value_2 not in dictionary:
            dictionary[value_2] = 0

        dictionary[value_2] += 1


def prepare_data(data_set: int, 
                 source: str = 'filtered_data',
                 streets_on: bool = False,
                 dump: bool = False,
                 path: str = '') -> Tuple[list, dict, Tuple[dict, int]]:
    '''Prepares the data for the plots by reading the filtered data 
    from a json file and returning the lines, districts and streets
    '''

    file_path = os.path.join(f'{source}', f'filtered_data_{data_set}.json')
    file = open(file_path)
    file_data = json.load(file)

    lines = []
    districts = {}
    streets = {}

    district_json = load_geojson('warszawa-dzielnice.geojson')

    for feature in district_json['features'][1:]:
        if feature['properties']['name'] not in districts:
            districts[feature['properties']['name'] + "_total"] = 0

    for info in file_data:
        if info['Velocity'] > SPEED_LIMIT:
            lines.append((info["Start_pos"], info["End_pos"]))

            increase_dict_values(
                districts,
                info["District_Start"],
                info["District_End"])

            if streets_on:
                if info["Street_1"] == "" or info["Street_2"] == "":
                    print("This data set is not suitable for street analysis")
                    return (lines, districts, None)

                increase_dict_values(
                    streets, info["Street_1"], info["Street_2"])

        # We keep the total number of data collected in each district
        if info["District_Start"] != info["District_End"]:
            districts[info["District_End"] + "_total"] += 1

        districts[info["District_Start"] + "_total"] += 1

    total = sum([count for district, count in districts.items()
                if district.endswith("_total")])

    file.close()
    
    if dump and path != '':
        try:
            with open(os.path.join(path, f'data_dump_{data_set}.json'), 'w') as file:
                json.dump((lines, districts, (streets, total)), file)
        except FileNotFoundError:
            print(f"ERROR: File {path}/data_dump_{data_set}.json does not exist")

    return (lines, districts, (streets, total))



import os
import json
import matplotlib.pyplot as plt
import numpy as np
from typing import Tuple
from .helper_functions import create_directory, load_geojson

# Set the font styles for the plots to ensure consistency and cohesion
fontstyle_annotation = {'fontsize': 9,
                        'fontweight': 'bold', 'color': '#14333D'}

fontstyle_small = {'fontsize': 8, 'color': '#14333D'}

fontstyle_title = {'fontsize': 13, 'fontweight': 'bold', 'color': '#14333D'}

# Colours for each of the districts
colours = ['#3943B7', '#83C5BE', '#0CF574', '#617073', '#E29578', '#BF9ACA',
           '#EB5160', '#FF7F11', '#2AB7CA', '#7BD389', '#521945', '#9B7EDE',
           '#76E7CD', '#29335C', '#AF125A', '#C6BF06', '#35A7FF', '#706C61']

SPEED_LIMIT = 50


def convert_data(data: list) -> list:
    '''Converts the data from seconds to minutes and
    puts it into a dictionary
    '''

    new_data = []

    for info in data:
        info = info // 60

        new_data.append(info)

    return new_data


def get_percentage(data: list) -> float:
    '''Calculates the percentage of busses that are late in a district'''

    late = 0

    for info in data:
        if info > 5:
            late += 1

    return (late / len(data)) * 100


def plot_district(district: str, data: list, plot: plt.Axes,
                  colour: str, order: str) -> None:
    '''Plots the number of minutes late for a specific district'''

    frequencies = {}
    data = convert_data(data)

    for number in data:
        if number in frequencies:
            frequencies[number] += 1
        else:
            frequencies[number] = 1

    x = list(frequencies.keys())
    y = list(frequencies.values())

    plot.set_title(f'{district} #{order}', fontstyle_annotation)
    plot.set_xlabel('Number of minutes late', fontstyle_small)
    plot.set_ylabel('Frequency', fontstyle_small)

    # Each district has a different colour from the colours array
    plot.bar(x, y, color=colour)


def plot_warsaw(data, plot: plt.Axes) -> None:
    '''Plots the number of minutes late for all of Warsaw'''

    frequencies = {}

    for district in data:
        for number in convert_data(data[district]):
            if number in frequencies:
                frequencies[number] += 1
            else:
                frequencies[number] = 1

    x = list(frequencies.keys())
    y = list(frequencies.values())

    plot.set_title('Warsaw', fontstyle_annotation)
    plot.set_xlabel('Number of minutes late', fontstyle_small)
    plot.set_ylabel('Frequency', fontstyle_small)

    plot.bar(x, y)


def plot_district_percentage(districts, data, plot: plt.Axes) -> None:
    '''Plots the percentage of busses late in each district in one plot'''

    percentages = {}
    district_number = 0

    for district in districts:
        if district not in data:
            continue

        if len(data[district]) == 0:  # Avoids division by zero
            continue

        percentages[district_number +
                    1] = get_percentage(convert_data(data[district]))

        # Each district is assigned a number from 1 to 18
        district_number += 1

    x = list(percentages.keys())
    y = list(percentages.values())

    plot.set_title(
        'The percentage of busses\nlate in each district',
        fontstyle_annotation)
    plot.set_xlabel('District number', fontstyle_small)
    plot.set_ylabel('Percentage', fontstyle_small)

    plot.bar(x, y, color=colours)


def configure_plot() -> Tuple[plt.Figure, plt.Axes]:
    '''Creates the appropriate figure and plot to plot the
    punctuality of Warsaw's busses
    '''

    figure, plot = plt.subplots(4, 5, figsize=(30, 20))
    figure.subplots_adjust(wspace=0.6, hspace=0.6)

    figure.suptitle(
        'The number of minutes late in Warsaw and all of Warsaw\'s districts',
        fontsize=16,
        fontweight='bold',
        fontdict=fontstyle_title,
        y=0.96)

    return figure, plot


def plot_lateness(data_set: int, 
                  src_path: str = 'lateness_data',
                  dst_path: str = 'plots') -> None:
    '''Plots the number of minutes late for all of Warsaw and
    all of Warsaw's districts
    '''

    # Creates a directory for the plots if it does not already exist
    create_directory(data_set)

    try:
        with open(f'{src_path}/lateness_{data_set}.json', 'r') as file:
            lateness_data = json.load(file)
    except FileNotFoundError:
        print(f"ERROR: File {src_path}/lateness_{data_set}.json \
              does not exist")
        return

    districts_file = load_geojson('warszawa-dzielnice.geojson')

    districts = sorted([district['properties']['name']
                       for district in districts_file['features'][1:]])

    figure, plot = configure_plot()

    iter = 0

    for i in range(20):
        row = i // 5
        col = i % 5

        if i == 7:
            plot_warsaw(lateness_data, plot[row][col])
            continue

        if i == 12:
            plot_district_percentage(districts, lateness_data, plot[row][col])
            continue

        if districts[iter] in lateness_data:
            plot_district(districts[iter],
                          lateness_data[districts[iter]],
                          plot[row][col],
                          colours[iter],
                          f'{iter + 1}')

        iter += 1

    plt.get_current_fig_manager().full_screen_toggle()
    plt.savefig(f'{dst_path}/plots_{data_set}/Bus_punctuality_plot_{data_set}.png')


def graph_speed_versus_delays(data: list, data_set: int, path) -> None:
    '''Graphs the relationship between the percentage of busses late and the
    percentage of busses speeding for the 10 busses with the highest
    delays to speeding ratio
    '''
    
    x = [f'{entry[0][0]}-{entry[0][1]}-{entry[0][2]}' for entry in data[:10]]
    y1 = [entry[1] for entry in data[:10]]
    y2 = [entry[2] for entry in data[:10]]

    fig, ax = plt.subplots(figsize=(10, 6))
    width = 0.4
    ax.bar(x, y1, label='Percentage Late', width=width)
    ax.bar([i + width for i in range(len(x))], y2,
           label='Speeding Percentage', width=width)

    ax.set_xlabel('Line-Brigade-Vehicle Number')
    ax.set_ylabel('Percentage')
    ax.set_title(
        'Percentage Late and Speeding Percentage for the 10 Busses Sorted by their Delays to Speeding Ratio')
    ax.legend()

    plt.xticks(rotation=45)
    plt.tight_layout()
    plt.savefig(
        f'{path}/plots_{data_set}/Speeding_versus_delays_plot_{data_set}.png')


def plot_speed_versus_delays(data_set: int, 
                             path: str = 'plots', 
                             filtered_path: str = 'filtered_data',
                             lateness_path: str = 'lateness_data') -> None:
    '''Plots the relationship between a vehicle going over the speed limit and 
    them being on time
    '''

    try:
        with open(f'{lateness_path}/lateness_lines_{data_set}.json', 'r') as f:
            lateness_data = json.load(f)
    except FileNotFoundError:
        print(f"ERROR: File {lateness_path}/lateness_lines_{data_set}.json \
              does not exist")
        return

    try:
        with open(f'{filtered_path}/filtered_data_{data_set}.json', 'r') as f:
            filtered_data = json.load(f)
    except FileNotFoundError:
        print(f"ERROR: File {filtered_path}/filtered_data_{data_set}.json \
              does not exist")
        return

    data = []

    for line in lateness_data:
        if lateness_data[line][1] == []:
            continue

        percentage_late = get_percentage(convert_data(lateness_data[line][1]))

        speeding = 0
        all_measurements = 0

        for info in filtered_data:
            if info['Line'] == lateness_data[line][0][0] and info['Brigade'] == lateness_data[
                    line][0][1] and info['Vehicle_number'] == lateness_data[line][0][2]:
                all_measurements += 1

                if info['Velocity'] > SPEED_LIMIT:
                    speeding += 1

        if all_measurements == 0:
            continue

        speeding_percentage = round(speeding / all_measurements * 100, 2)

        if speeding_percentage == 0:
            continue

        data.append(
            (lateness_data[line][0],
             percentage_late,
             speeding_percentage,
             percentage_late /
             speeding_percentage))

    data.sort(key=lambda x: x[3], reverse=True)

    graph_speed_versus_delays(data, data_set, path)



import setuptools

with open("README.md", "r", encoding="UTF-8") as fh:
    description_2 = fh.read()

setuptools.setup(
    name="bus_project_NJ",
        version="0.0.3",
        author="Natalia J.",
        author_email="njunkiert@gmail.com",
        description="A package for analysing public transport data",
        long_description=description_2,
        long_description_content_type="text/markdown",
        url='https://gitlab.mimuw.edu.pl/nj448267/python_project_nj',
        packages=setuptools.find_packages(),
        package_data={'bus_project_NJ': ['/src/*geojson']},
        classifiers=[
            "Programming Language :: Python :: 3", 
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent"
        ],
        python_requires='>=3.6'
)


import json
import os
from .helper_functions import fetch_data

# Set the url for UM Warszawa
url = 'https://api.um.warszawa.pl/api/action/dbtimetable_get'
api_key = ""


def get_bus_lines(path: str = 'bus_data') -> list:
    '''Returns a list of bus lines in Warsaw'''

    with open(f'{path}/bus_routes.json', 'r') as file:
        bus_routes = json.load(file)

    lines = list(bus_routes['result'])

    file.close()

    return lines


def format_response(response, bus_schedule: dict, key: str) -> None:
    ''' Formats response into a friendlier format and saves it in the
    line's schedule dictionary
    '''

    for info in response['result']:
        formatted_data = dict()

        for dic in info["values"]:
            formatted_data[dic["key"]] = dic["value"]

        brigade = formatted_data['brygada']

        if brigade not in bus_schedule:
            bus_schedule[brigade] = dict()

        if key not in bus_schedule[brigade]:
            bus_schedule[brigade][key] = list()

        bus_schedule[brigade][key].append(formatted_data['czas'])


def get_line_schedule(line: str, path: str) -> None:
    '''Collects the bus schedule for a specific bus line and
    saves it in a json file
    '''

    bus_line_dir = os.path.join(f'{path}/bus_lines')

    try:
        with open(f'{path}/bus_routes.json', 'r') as file:
            bus_routes = json.load(file)
    except FileNotFoundError:
        print("ERROR: Bus routes file not found. Please run \
              fetch_bus_routes.py first.")
        return

    bus_schedule = dict()

    for route in bus_routes['result'][line]:
        for stop in bus_routes['result'][line][route]:
            info = bus_routes['result'][line][route][stop]

            # Parameters for the request
            params = {
                'apikey': api_key,
                'id': 'e923fa0e-d96c-43f9-ae6e-60518c9f3238',
                'busstopId': info['nr_zespolu'],
                'busstopNr': info['nr_przystanku'],
                'line': line}

            response = fetch_data(params, url)

            # If the request was unsuccessful, skip to the next stop
            if response is None:
                continue

            key = f"{info['nr_zespolu']}-{info['nr_przystanku']}"

            format_response(response, bus_schedule, key)

    with open(os.path.join(bus_line_dir, 
                           f'bus{line}_schedule.json'), 'w') as file:
        json.dump(bus_schedule, file)


def collect_bus_schedules(apikey: str, path: str = 'bus_data') -> None:
    '''Collects the bus schedules for all the bus lines in 
    Warsaw and saves them in a json file
    '''

    global api_key
    api_key = apikey

    bus_lines_dir = os.path.join(f'{path}', 'bus_lines')

    # Create the directory if it doesn't exist already
    if not os.path.exists(bus_lines_dir):
        os.makedirs(bus_lines_dir)

    bus_lines = get_bus_lines(path)
    
    for bus in bus_lines:
        print(f'Collecting data for line {bus}')
        get_line_schedule(bus, path)

